/* $Id: atari_dc.c,v 1.47 2003/01/17 00:07:39 chris Exp $
 *
 * SEGA Dreamcast support using KallistiOS (http://cadcdev.sourceforge.net)
 * (c) 2002,2003 Christian Groessler (chris@groessler.org)
 */

#include <stdarg.h>
#include <stdlib.h>
#include <kos.h>
#include <sys/types.h>
#include "atari.h"
#include "config.h"
#include "sio.h"
#include "binload.h"
#include "input.h"
#include "rt-config.h"
#include "colours.h"
#include "ui.h"
#include "time.h"
#include "kb_ui.h"
#include <dc/g2bus.h>
#include <arm/aica_cmd_iface.h>
#include <dc/sound/sound.h>
#include "statesav.h"
#include "sound.h"
#define _TYPEDEF_H   /* to prevent pokeysnd.h to create uint32 type #defines */
#include "pokeysnd.h"

#define REPEAT_DELAY 5000
#define REPEAT_INI_DELAY (5 * REPEAT_DELAY)
/*#define DEBUG*/
/*#define SPEED_CHECK*/
/*#define MY_RELEASE*/
#ifndef DIRTYRECT
#error need DIRTYRECT
#endif

void Atari_DisplayScreen(UBYTE *);
int stat(const char *, struct stat *);
static void calc_palette(void);
static int check_tray_open(void);
static void autostart(void);
static void controller_update(void);
static void dc_atari_sync(void);
static void dc_sound_init(void);

static unsigned char *atari_screen_backup;
static unsigned char *atari_screen_backup2;
static uint8 mcont[4];
static cont_cond_t mcond[4];
static uint8 mkb;
static int num_cont;   /* # of controllers found */
static int su_first_call = TRUE;
static int b_ui_leave = FALSE;
static int in_kbui = FALSE;
static int open_tray = FALSE;
static int tray_closed = TRUE;
int emulate_paddles = FALSE;
int glob_snd_ena = TRUE;
int db_mode = TRUE;
int in_fnedit = FALSE;
int screen_tv_mode;    /* mode of connected tv set */

int x_ovr = FALSE, y_ovr = FALSE, b_ovr = FALSE;
int ovr_inject_key = AKEY_NONE;
int x_key = AKEY_NONE;
int y_key = AKEY_NONE;
int b_key = AKEY_NONE;

extern char curr_disk_dir[];
extern char curr_cart_dir[];
extern char curr_exe_dir[];
extern char curr_tape_dir[];

#define DSPRATE 22050

static vid_mode_t mymode_pal = {
	DM_320x240,
	320, 240,
	VID_PIXELDOUBLE|VID_LINEDOUBLE|VID_PAL,
	CT_ANY,
	PM_RGB565,
	313, 857,
	0xA4, 0x35, /*0x2d,*/
	0x15, 312,
	141, 843,
	24, 313*2,
	0, 1,
	{ 0, 0, 0, 0 }
};
static vid_mode_t mymode_ntsc = {
	DM_320x240,
	320, 240,			      /* width, height in pixels */
	VID_PIXELDOUBLE|VID_LINEDOUBLE,	      /* flags */
	CT_ANY,				      /* cable type */
	PM_RGB565,			      /* pixel mode */
	262, 857,			      /* scanlines/frame, clocks/scanline */
	0xA4, 0x18,			      /* bitmap window position x, y */
	0x15, 261,			      /* scanline interrupt positions */
	141, 843,			      /* border x start + stop */
	24, 263*2,			      /* border y start + stop */
	0, 1,
	{ 0, 0, 0, 0 }
};

void update_vidmode(void)
{
	vid_mode_t mymode;

	/* we need to copy it since vid_set_mode_ex() modifies it */
	switch(screen_tv_mode) {
	case TV_NTSC:
		memcpy(&mymode, &mymode_ntsc, sizeof(vid_mode_t));
		break;
	case TV_PAL:
	default:
		memcpy(&mymode, &mymode_pal, sizeof(vid_mode_t));
		break;
	}
	mymode.cable_type = vid_check_cable();
	vid_set_mode_ex(&mymode);
}

void Atari_Initialise(int *argc, char *argv[])
{
	static int fc = TRUE;

	if (! fc) return; else fc = FALSE;

	screen_tv_mode = tv_mode;

#ifndef DEBUG
	/* Bother us with output only if something died */
	dbglog_set_level(DBG_DEAD);
#endif

	/* Set the video mode */
	update_vidmode();
	calc_palette();
}

#ifdef SUPPORTS_ATARI_CONFIGINIT

void Atari_ConfigInit(void)  /* not currently used */
{
	strcpy(atari_osa_filename, "/cd/atariosa.rom");
	strcpy(atari_osb_filename, "/cd/atariosb.rom");
	strcpy(atari_xlxe_filename, "/cd/atarixl.rom");
	strcpy(atari_basic_filename, "/cd/ataribas.rom");
	strcpy(atari_5200_filename, "/cd/5200.rom");
	strcpy(atari_disk_dirs[0], "/cd/disks");
	disk_directories = 1;
	strcpy(atari_rom_dir, "/cd/cartridges");
	strcpy(atari_h1_dir, "/cd");
	strcpy(atari_h2_dir, "/cd");
	strcpy(atari_h3_dir, "/cd");
	strcpy(atari_h4_dir, "/cd");
	strcpy(atari_exe_dir, "/cd/files");
	strcpy(atari_state_dir, "/vmu");
}

#endif

int Atari_Exit(int run_monitor)
{
	arch_reboot();
	return(0);  /* not reached */
}

/*
 * update the in-core status of the controller buttons/settings
 */
static void controller_update(void)
{
	int i;
	for (i=0; i<num_cont; i++) {
		if (cont_get_cond(mcont[i], &mcond[i])) {
#ifdef DEBUG
			printf("controller_update: error getting controller status\n");
#endif
		}
	}
}

static int consol_keys(void)
{
	cont_cond_t *cond = &mcond[0];

	if (! (cond->buttons & CONT_START)) {
		if (ui_is_active)
			arch_reboot();
		else {
			if (! (cond->buttons & CONT_X))
				return(AKEY_COLDSTART);
			else
				return(AKEY_WARMSTART);
		}
	}

	if (machine_type != MACHINE_5200) {
		if (b_ovr && !ui_is_active) {  /* !ui_is_active b/c B is also the esc key (with rtrig) */
			if (b_key != AKEY_NONE && !(cond->buttons & CONT_B))
				ovr_inject_key = b_key;
		}
		else {
			if (! (cond->buttons & CONT_B)) {
				if (! b_ui_leave) {
					key_consol &= ~CONSOL_START;
				}
				else {
					key_consol |= CONSOL_START;
				}
			}
			else {
				b_ui_leave = FALSE;
				key_consol |= CONSOL_START;
			}
		}

		if (y_ovr) {
			if (y_key != AKEY_NONE && !(cond->buttons & CONT_Y))
				ovr_inject_key = y_key;
		}
		else {
			if (! (cond->buttons & CONT_Y)) {
				key_consol &= ~CONSOL_SELECT;
			}
			else {
				key_consol |= CONSOL_SELECT;
			}
		}

		if (x_ovr) {
			if (x_key != AKEY_NONE && !(cond->buttons & CONT_X))
				ovr_inject_key = x_key;
		}
		else {
			if (! (cond->buttons & CONT_X)) {
				key_consol &= ~CONSOL_OPTION;
			}
			else {
				key_consol |= CONSOL_OPTION;
			}
		}
	}
	else {
		/* @@@ *_ovr TODO @@@ */

		if (! (cond->buttons & CONT_X)) {  /* 2nd action button */
			key_shift = 1;
		}
		else {
			key_shift = 0;
		}
		if (! ui_is_active) {
			if (! (cond->buttons & CONT_B)) {  /* testtest */
				if (! b_ui_leave) {
					return(AKEY_4);	  /* ??? at least on the dc kb "4" starts some games?? */
				}
			}
			else {
				b_ui_leave = FALSE;
			}
			if (! (cond->buttons & CONT_Y)) {  /* testtest */
				return(AKEY_5200_START);
			}
		}
	}
	return(AKEY_NONE);
}

int get_emkey(UBYTE *title)
{
	int keycode;

	controller_update();
	in_kbui = TRUE;
	memcpy(atari_screen_backup, atari_screen, ATARI_HEIGHT * ATARI_WIDTH);
	kb_ui((UBYTE *)atari_screen, title, KB_SELECT);
	ui_is_active = TRUE;
	memcpy(atari_screen, atari_screen_backup, ATARI_HEIGHT * ATARI_WIDTH);
	entire_screen_dirty();
	in_kbui = FALSE;

	if (inject_key != AKEY_NONE) {
		keycode = inject_key;
		inject_key = AKEY_NONE;
		return(keycode);
	}
	else {
		return(AKEY_NONE);
	}

}

/*
 * do some basic keyboard emulation for the controller,
 * so that the emulator menu can be used without keyboard
 * (basically for selecting bin files/disk images)
 */
static int controller_kb(void)
{
	static int prev_up = FALSE, prev_down = FALSE, prev_a = FALSE,
		prev_r = FALSE, prev_left = FALSE, prev_right = FALSE,
		prev_b = FALSE;
	static int repdelay = REPEAT_DELAY;
	cont_cond_t *cond = &mcond[0];

	if (! num_cont) return(AKEY_NONE);  /* no controller present */

	repdelay--;
	if (repdelay < 0) repdelay = REPEAT_DELAY;

	if (!ui_is_active && (cond->ltrig > 250 || !(cond->buttons & CONT_Z))) {
		return(AKEY_UI);
	}
#ifdef KB_UI
	if (!ui_is_active && (cond->rtrig > 250 || !(cond->buttons & CONT_C))) {
		controller_update();
		return(AKEY_KEYB);
	}
	/* provide keyboard emulation to enter save file name */
	if (ui_is_active && in_fnedit && !in_kbui && (cond->rtrig > 250 || !(cond->buttons & CONT_C))) {
		controller_update();
		in_kbui = TRUE;
		memcpy(atari_screen_backup, atari_screen, ATARI_HEIGHT * ATARI_WIDTH);
		kb_ui((UBYTE *)atari_screen, NULL, 0);
		ui_is_active = TRUE;
		memcpy(atari_screen, atari_screen_backup, ATARI_HEIGHT * ATARI_WIDTH);
		entire_screen_dirty();
		in_kbui = FALSE;
		if (inject_key != AKEY_NONE) {
			int keycode;
			keycode = inject_key;
			inject_key = AKEY_NONE;
			return(keycode);
		}
		else {
			return(AKEY_NONE);
		}
	}
#endif

	if (ui_is_active) {
		if (! (cond->buttons & CONT_DPAD_UP)) {
			prev_down = FALSE;
			if (! prev_up) {
				repdelay = REPEAT_INI_DELAY;
				prev_up = 1;
				return(AKEY_UP);
			}
			else {
				if (! repdelay) {
					return(AKEY_UP);
				}
			}
		}
		else {
			prev_up = FALSE;
		}

		if (! (cond->buttons & CONT_DPAD_DOWN)) {
			prev_up = FALSE;
			if (! prev_down) {
				repdelay = REPEAT_INI_DELAY;
				prev_down = TRUE;
				return(AKEY_DOWN);
			}
			else {
				if (! repdelay) {
					return(AKEY_DOWN);
				}
			}
		}
		else {
			prev_down = FALSE;
		}

		if (! (cond->buttons & CONT_DPAD_LEFT)) {
			prev_right = FALSE;
			if (! prev_left) {
				repdelay = REPEAT_INI_DELAY;
				prev_left = TRUE;
				return(AKEY_LEFT);
			}
			else {
				if (! repdelay) {
					return(AKEY_LEFT);
				}
			}
		}
		else {
			prev_left = FALSE;
		}

		if (! (cond->buttons & CONT_DPAD_RIGHT)) {
			prev_left = FALSE;
			if (! prev_right) {
				repdelay = REPEAT_INI_DELAY;
				prev_right = TRUE;
				return(AKEY_RIGHT);
			}
			else {
				if (! repdelay) {
					return(AKEY_RIGHT);
				}
			}
		}
		else {
			prev_right = FALSE;
		}

		if (! (cond->buttons & CONT_A)) {
			if (! prev_a) {
				prev_a = TRUE;
				return(AKEY_RETURN);
			}
		}
		else {
			prev_a = FALSE;
		}

		if (! (cond->buttons & CONT_B)) {
			if (! prev_b) {
				prev_b = TRUE;
				b_ui_leave = TRUE;   /* B must be released again */
				return(AKEY_ESCAPE);
			}
		}
		else {
			prev_b = FALSE;
		}

		if (cond->rtrig > 250 || !(cond->buttons & CONT_C)) {
			if (! prev_r) {
				prev_r = TRUE;
				return(AKEY_ESCAPE);
			}
		}
		else {
			prev_r = FALSE;
		}
	}
	return(AKEY_NONE);
}

int Atari_Keyboard(void)
{
	static int old_open_tray = FALSE;
	int open_tray;
	int keycode;

	/* let's do this here... */
	open_tray = check_tray_open();
	if (open_tray != old_open_tray) {
		old_open_tray = open_tray;
		if (open_tray == TRUE) {
#ifdef DEBUG
			printf("XXX TRAY OPEN!\n");
#endif
			tray_closed = FALSE;
			return(AKEY_UI);
		}
		else {
#ifdef DEBUG
			printf("XXX TRAY CLOSE!\n");
#endif
			tray_closed = TRUE;
			cdrom_init();
			iso_ioctl(0, NULL, 0); /* @@@ check what this is in new KOS version ! */
			strcpy(curr_disk_dir, "/");
			strcpy(curr_cart_dir, "/");
			strcpy(curr_exe_dir, "/");
			strcpy(curr_tape_dir, "/");
		}
	}

	if (ui_is_active) controller_update();

	if (num_cont && (keycode = consol_keys()) != AKEY_NONE) return(keycode);

#ifdef KB_UI
	if (inject_key != AKEY_NONE) {
		keycode = inject_key;
		inject_key = AKEY_NONE;
		switch(keycode) {
		case AKEY_OPTION:
			key_consol &= (~CONSOL_OPTION);
			keycode = AKEY_NONE;
			break;
		case AKEY_SELECT:
			key_consol &= (~CONSOL_SELECT);
			keycode = AKEY_NONE;
			break;
		case AKEY_START:
			key_consol &= (~CONSOL_START);
			keycode = AKEY_NONE;
			break;
		}
		return(keycode);
	}
#endif
	if (ovr_inject_key != AKEY_NONE) {
		keycode = ovr_inject_key;
		ovr_inject_key = AKEY_NONE;
		return(keycode);
	}

	keycode = controller_kb();
	if (keycode != AKEY_NONE) return(keycode);

	keycode = kbd_get_key();
	if (keycode == -1) return(AKEY_NONE);

#ifdef DEBUG
	printf("DC key: %04X (%c)\n", keycode, keycode > 32 && keycode < 127 ? keycode : '.');
#endif

	switch (keycode) {

	/* OPTION / SELECT / START keys */
	/*key_consol = CONSOL_NONE;  -- already set in consol_keys... */
	case 0x3b00:
		key_consol &= (~CONSOL_OPTION);
		return(AKEY_NONE);
	case 0x3c00:
		key_consol &= (~CONSOL_SELECT);
		return(AKEY_NONE);
	case 0x3d00:
		key_consol &= (~CONSOL_START);
		return(AKEY_NONE);

	case 0x1b:  /* ESC */
		keycode = AKEY_ESCAPE;
		break;
	case '`':
		keycode = AKEY_CAPSTOGGLE;
		break;
	case '!':
		keycode = AKEY_EXCLAMATION;
		break;
	case '"':
		keycode = AKEY_DBLQUOTE;
		break;
	case '#':
		keycode = AKEY_HASH;
		break;
	case '$':
		keycode = AKEY_DOLLAR;
		break;
	case '%':
		keycode = AKEY_PERCENT;
		break;
	case '&':
		keycode = AKEY_AMPERSAND;
		break;
	case '\'':
		keycode = AKEY_QUOTE;
		break;
	case '@':
		keycode = AKEY_AT;
		break;
	case '(':
		keycode = AKEY_PARENLEFT;
		break;
	case ')':
		keycode = AKEY_PARENRIGHT;
		break;
	case '[':
		keycode = AKEY_BRACKETLEFT;
		break;
	case ']':
		keycode = AKEY_BRACKETRIGHT;
		break;
	case '<':
		keycode = AKEY_LESS;
		break;
	case '>':
		keycode = AKEY_GREATER;
		break;
	case '=':
		keycode = AKEY_EQUAL;
		break;
	case '?':
		keycode = AKEY_QUESTION;
		break;
	case '-':
		keycode = AKEY_MINUS;
		break;
	case '+':
		keycode = AKEY_PLUS;
		break;
	case '*':
		keycode = AKEY_ASTERISK;
		break;
	case '/':
		keycode = AKEY_SLASH;
		break;
	case ':':
		keycode = AKEY_COLON;
		break;
	case ';':
		keycode = AKEY_SEMICOLON;
		break;
	case ',':
		keycode = AKEY_COMMA;
		break;
	case '.':
		keycode = AKEY_FULLSTOP;
		break;
	case '_':
		keycode = AKEY_UNDERSCORE;
		break;
	case '^':
		keycode = AKEY_CIRCUMFLEX;
		break;
	case '\\':
		keycode = AKEY_BACKSLASH;
		break;
	case '|':
		keycode = AKEY_BAR;
		break;
	case ' ':
		keycode = AKEY_SPACE;
		break;
	case '0':
		keycode = AKEY_0;
		break;
	case '1':
		keycode = AKEY_1;
		break;
	case '2':
		keycode = AKEY_2;
		break;
	case '3':
		keycode = AKEY_3;
		break;
	case '4':
		keycode = AKEY_4;
		break;
	case '5':
		keycode = AKEY_5;
		break;
	case '6':
		keycode = AKEY_6;
		break;
	case '7':
		keycode = AKEY_7;
		break;
	case '8':
		keycode = AKEY_8;
		break;
	case '9':
		keycode = AKEY_9;
		break;
	case 'a':
		keycode = AKEY_a;
		break;
	case 'b':
		keycode = AKEY_b;
		break;
	case 'c':
		keycode = AKEY_c;
		break;
	case 'd':
		keycode = AKEY_d;
		break;
	case 'e':
		keycode = AKEY_e;
		break;
	case 'f':
		keycode = AKEY_f;
		break;
	case 'g':
		keycode = AKEY_g;
		break;
	case 'h':
		keycode = AKEY_h;
		break;
	case 'i':
		keycode = AKEY_i;
		break;
	case 'j':
		keycode = AKEY_j;
		break;
	case 'k':
		keycode = AKEY_k;
		break;
	case 'l':
		keycode = AKEY_l;
		break;
	case 'm':
		keycode = AKEY_m;
		break;
	case 'n':
		keycode = AKEY_n;
		break;
	case 'o':
		keycode = AKEY_o;
		break;
	case 'p':
		keycode = AKEY_p;
		break;
	case 'q':
		keycode = AKEY_q;
		break;
	case 'r':
		keycode = AKEY_r;
		break;
	case 's':
		keycode = AKEY_s;
		break;
	case 't':
		keycode = AKEY_t;
		break;
	case 'u':
		keycode = AKEY_u;
		break;
	case 'v':
		keycode = AKEY_v;
		break;
	case 'w':
		keycode = AKEY_w;
		break;
	case 'x':
		keycode = AKEY_x;
		break;
	case 'y':
		keycode = AKEY_y;
		break;
	case 'z':
		keycode = AKEY_z;
		break;
	case 'A':
		keycode = AKEY_A;
		break;
	case 'B':
		keycode = AKEY_B;
		break;
	case 'C':
		keycode = AKEY_C;
		break;
	case 'D':
		keycode = AKEY_D;
		break;
	case 'E':
		keycode = AKEY_E;
		break;
	case 'F':
		keycode = AKEY_F;
		break;
	case 'G':
		keycode = AKEY_G;
		break;
	case 'H':
		keycode = AKEY_H;
		break;
	case 'I':
		keycode = AKEY_I;
		break;
	case 'J':
		keycode = AKEY_J;
		break;
	case 'K':
		keycode = AKEY_K;
		break;
	case 'L':
		keycode = AKEY_L;
		break;
	case 'M':
		keycode = AKEY_M;
		break;
	case 'N':
		keycode = AKEY_N;
		break;
	case 'O':
		keycode = AKEY_O;
		break;
	case 'P':
		keycode = AKEY_P;
		break;
	case 'Q':
		keycode = AKEY_Q;
		break;
	case 'R':
		keycode = AKEY_R;
		break;
	case 'S':
		keycode = AKEY_S;
		break;
	case 'T':
		keycode = AKEY_T;
		break;
	case 'U':
		keycode = AKEY_U;
		break;
	case 'V':
		keycode = AKEY_V;
		break;
	case 'W':
		keycode = AKEY_W;
		break;
	case 'X':
		keycode = AKEY_X;
		break;
	case 'Y':
		keycode = AKEY_Y;
		break;
	case 'Z':
		keycode = AKEY_Z;
		break;

	case 0x3a00:  /* F1 */
		keycode = AKEY_UI;
		break;

	case 0x3e00:  /* F5 */
		keycode = AKEY_COLDSTART;
		break;

	case 0x4500:  /* F12 */
		arch_reboot();

	/* cursor keys */
	case 0x5200:
		keycode = AKEY_UP;
		break;
	case 0x5100:
		keycode = AKEY_DOWN;
		break;
	case 0x5000:
		keycode = AKEY_LEFT;
		break;
	case 0x4f00:
		keycode = AKEY_RIGHT;
		break;

	case 0x6500:  /* S3 */
		keycode = AKEY_ATARI;
		break;

	case 0x4a00:  /* Home */
		keycode = AKEY_CLEAR;
		break;
	case 0x4d00:  /* End */
		keycode = AKEY_HELP;
		break;
	case 0x4800:  /* Break/Pause */
		keycode = AKEY_BREAK;
		break;

	case 0x4c00:  /* Del */
		if (key_shift)
			keycode = AKEY_DELETE_LINE;
		else
			keycode |= AKEY_DELETE_CHAR;
		break;
	case 0x4900:  /* Ins */
		if (key_shift)
			keycode = AKEY_INSERT_LINE;
		else
			keycode |= AKEY_INSERT_CHAR;
		break;

	case 127:
	case 8:
		keycode = AKEY_BACKSPACE;
		break;
	case 13:
	case 10:
		keycode = AKEY_RETURN;
		break;
	default:
		keycode = AKEY_NONE;
		break;
	}
	return keycode;
}

int Atari_PORT(int num)
{
	cont_cond_t *cond;
	int retval = 0xff;
	int cix = 0;

	if (num) {
		if (machine_type != MACHINE_OSA && machine_type != MACHINE_OSB) return(retval);
		cix = 2;   /* js #2 and #3 */
	}

	if (num_cont < 1 + cix) return(retval);	 /* no controller present */

	cond = &mcond[cix];
	if (! emulate_paddles) {
		if (! (cond->buttons & CONT_DPAD_UP)) {
			retval &= 0xfe;
		}
		if (! (cond->buttons & CONT_DPAD_DOWN)) {
			retval &= 0xfd;
		}
		if (! (cond->buttons & CONT_DPAD_LEFT)) {
			retval &= 0xfb;
		}
		if (! (cond->buttons & CONT_DPAD_RIGHT)) {
			retval &= 0xf7;
		}
		/* if joypad not used, try the joystick instead */
		if (retval == 0xff) {
			if (cond->joyx > 127 + 64) {  /* right */
				retval &= 0xf7;
			}
			if (cond->joyx < 127 - 64) {  /* left */
				retval &= 0xfb;
			}
			if (cond->joyy > 127 + 64) {  /* down */
				retval &= 0xfd;
			}
			if (cond->joyy < 127 - 64) {  /* up */
				retval &= 0xfe;
			}
		}
		if (num_cont > 1 + cix) {
			cond = &mcond[cix+1];
			if (! (cond->buttons & CONT_DPAD_UP)) {
				retval &= 0xef;
			}
			if (! (cond->buttons & CONT_DPAD_DOWN)) {
				retval &= 0xdf;
			}
			if (! (cond->buttons & CONT_DPAD_LEFT)) {
				retval &= 0xbf;
			}
			if (! (cond->buttons & CONT_DPAD_RIGHT)) {
				retval &= 0x7f;
			}
			/* if joypad not used, try the joystick instead */
			if ((retval & 0xf0) == 0xf0) {
				if (cond->joyx > 127 + 64) {  /* right */
					retval &= 0x7f;
				}
				if (cond->joyx < 127 - 64) {  /* left */
					retval &= 0xbf;
				}
				if (cond->joyy > 127 + 64) {  /* down */
					retval &= 0xdf;
				}
				if (cond->joyy < 127 - 64) {  /* up */
					retval &= 0xef;
				}
			}
		}
	}
	else {	/* emulate paddles */
		/* 1st paddle trigger */
		if (! (cond->buttons & CONT_A)) {
			retval &= (cix > 1) ? 0xbf : 0xfb;
		}
		if (num_cont > 1 + cix) {
			cond = &mcond[cix+1];

			/* 2nd paddle trigger */
			if (! (cond->buttons & CONT_A)) {
				retval &= (cix > 1) ? 0x7f : 0xf7;
			}
		}
	}

	return(retval);
}

int Atari_POT(int num)
{
	int val;
	cont_cond_t *cond;

	if (machine_type != MACHINE_5200) {
		if (emulate_paddles) {
			if (num + 1 > num_cont) return(228);
			if (machine_type != MACHINE_OSA && machine_type != MACHINE_OSB && num > 3) return(228);

			cond = &mcond[num];
			val = cond->joyx;
			val = val * 228 / 255;
			if (val > 227) return(1);
			return(228 - val);
		}
		else {
			return(228);
		}
	}
	else {	/* 5200 version:
		 *
		 * num / 2: which controller
		 * num & 1 == 0: x axis
		 * num & 1 == 1: y axis
		 */

		if (num / 2 + 1 > num_cont) return(joy_5200_center);

		cond = &mcond[num / 2];
		val = (num & 1) ? cond->joyy : cond->joyx;

		/* normalize into 5200 range */
		if (val == 127) return(joy_5200_center);
		if (val < 127) {
			/*val -= joy_5200_min;*/
			val = val * (joy_5200_center - joy_5200_min) / 127;
			return(val + joy_5200_min);
		}
		else {
			val = val * joy_5200_max / 255;
			if (val < joy_5200_center)
				val = joy_5200_center;
			return(val);
		}
	}
}

int Atari_TRIG(int num)
{
	cont_cond_t *cond;

	if (num + 1 > num_cont) return(1);  /* no controller present */
	if (machine_type != MACHINE_OSA && machine_type != MACHINE_OSB && num > 1) return(1);

	cond = &mcond[num];
	if (! (cond->buttons & CONT_A)) {
		return(0);
	}

	return(1);
}

static uint16 mypal[256];

static void calc_palette(void)
{
	int i;
	int rgb;
	int red;
	int green;
	int blue;
	uint16 pv;

	for (i = 0; i < 256; i++) {
		rgb = colortable[i];

		red = (rgb & 0x00ff0000) >> 16;
		green = (rgb & 0x0000ff00) >> 8;
		blue = (rgb & 0x000000ff) ;

		pv = ((red >> 3) << 11) |
			((green >> 2) << 5) |
			(blue >> 3);

		mypal[i] = pv;
	}
}

static void vbl_wait(void)
{
	static int last_cntr = 0;

	while (maple_state.vbl_cntr == last_cntr) ;
	last_cntr = maple_state.vbl_cntr;
}

static UBYTE old_sd[2][ATARI_HEIGHT * ATARI_WIDTH/8];

static void Atari_DisplayScreen_doubleb(UBYTE *screen)
{
	static unsigned int sbase = 0;	/* screen base of current frame */
	unsigned int x, y, m, j;
	uint16 *vram;
	UBYTE *osd, *nsd;  /* old screen-dirty, new screen-dirty */

#ifdef SPEED_CHECK
	vid_border_color(0, 0, 0);
#endif
	if (sbase) {
		sbase = 0;
		osd = old_sd[0];
		nsd = old_sd[1];
	}
	else {
		sbase = 1024 * 768 * 4;
		osd = old_sd[1];
		nsd = old_sd[0];
	}

	vram = (uint16*)(0xA5000000 + sbase) + (vid_mode->height - 240) / 2 * vid_mode->width;
	vram += (vid_mode->height - 240) / 2 * vid_mode->width;
	vram += (vid_mode->width - 320) / 2 - (ATARI_WIDTH - 320) / 2 ;

	for (m=0, x=0, y=0; m < ATARI_HEIGHT * ATARI_WIDTH / 8; m++) {
		nsd[m] = screen_dirty[m];  /* remember for other page */
		if (/*1 ||*/ screen_dirty[m] || osd[m]) {
			if (x - (ATARI_WIDTH - 320) / 2 <= 320 - 8) {
				/* Draw eight pixels (x,y), (x+1,y),...,(x+7,y) here */
				for (j=0; j<8; j++) {
					*(vram + x + j) = mypal[*(screen + x + j)];
				}
				screen_dirty[m] = 0;
			}
		}
		x += 8;
		if (x >= ATARI_WIDTH) {
			x = 0;
			y++;
			vram += vid_mode->width;
			screen += ATARI_WIDTH;
		}
	}
	vbl_wait();
	vid_set_start(sbase);
#ifdef SPEED_CHECK
	vid_border_color(255, 255, 255);
#endif
}


static void Atari_DisplayScreen_singleb(UBYTE *screen)
{
	unsigned int x, y, m, j;
	uint16 *vram = vram_s;

	/* point to current line of atari screen in display memory */
	vram += (vid_mode->height - 240) / 2 * vid_mode->width;

	/* adjust left column to screen resolution */
	vram += (vid_mode->width - 320) / 2 - (ATARI_WIDTH - 320) / 2 ;

	for (m=0, x=0, y=0; m < ATARI_HEIGHT * ATARI_WIDTH / 8; m++) {
		if (screen_dirty[m]) {
			if (x - (ATARI_WIDTH - 320) / 2 <= 320 - 8) {
				/* Draw eight pixels (x,y), (x+1,y),...,(x+7,y) here */
				for (j=0; j<8; j++) {
					*(vram + x + j) = mypal[*(screen + x + j)];
				}
				screen_dirty[m] = 0;
			}
		}
		x += 8;
		if (x >= ATARI_WIDTH) {
			x = 0;
			y++;
			vram += vid_mode->width;
			screen += ATARI_WIDTH;
		}
	}
	/*vbl_wait();*/
}

static void (*screen_updater)(UBYTE *screen) = Atari_DisplayScreen_doubleb;

void Atari_DisplayScreen(UBYTE *screen)
{
	screen_updater(screen);
}

void update_screen_updater(void)
{
	if (db_mode) {
		screen_updater = Atari_DisplayScreen_doubleb;
	}
	else {
		screen_updater = Atari_DisplayScreen_singleb;
	}
}

static void dc_controller_init(void)
{
	int p, u;

	for (p=0; p<MAPLE_PORT_COUNT; p++) {
		for (u=0; u<MAPLE_UNIT_COUNT; u++) {
			if (maple_device_func(p, u) & MAPLE_FUNC_CONTROLLER) {
				mcont[num_cont++] = maple_addr(p, u);
#ifdef DEBUG
				printf("dc_controller_init: found controller %d\n", mcont[num_cont-1]);
#endif
			}
		}
	}

	mkb = maple_first_kb();

#ifdef DEBUG
	printf("dc_controller_init: controller%sfound, keyboard%sfound\n",
	       mcont[0] ? " " : " not ", mkb ? " " : " not ");
#endif
	return;
}

/*
 * if state dir is empty or ".", setup the state dir to point
 * to an existing VMU
 */
static void setup_state_dir(void)
{
	int p, u;
	unsigned char name[4];

	name[0] = 0;
	if (!*atari_state_dir || !strcmp(atari_state_dir, ".")) {
		/* find first vmu */
		for (p=0; p<MAPLE_PORT_COUNT && !name[0]; p++) {
			for (u=0; u<MAPLE_UNIT_COUNT && !name[0]; u++) {
				if (maple_device_func(p, u) & MAPLE_FUNC_MEMCARD) {
					sprintf(name, "%c%c", 'a' + p, '0' + u);
				}
			}
		}
		if (name) {
			strcpy(atari_state_dir, "/vmu/");
			strcat(atari_state_dir, name);
		}
	}
}

#if defined(MY_RELEASE) || defined(HZ_TEST)
static void cls(int xres, int yres)
{
	int x, y;
	for(x = 0; x < xres; x++)
		for(y = 0; y < yres; y++)
			vram_s[xres*y + x] = 0;
}

static void wait_a(void)
{
	cont_cond_t cond;

	do {
		if (cont_get_cond(mcont[0], &cond)) {
			return;
		}

	} while (cond.buttons & CONT_A);
}
#endif

#ifdef HZ_TEST
void do_hz_test(void)
{
	uint32 s, ms, s2, ms2, z;
	char buffer[80];

	cls(320,240);
	bfont_draw_str(vram_s+64*320+5, 320, 0, "chk dsply hz.. 500 vbls");
	timer_ms_gettime(&s, &ms);
	for (z=0; z<500; z++) {
		vbl_wait();
		/*vid_waitvbl();*/
	}
	timer_ms_gettime(&s2, &ms2);
	sprintf(buffer, "start time: %lu.%03lu\n", s, ms);
	bfont_draw_str(vram_s+89*320+5, 320, 0, buffer);
	sprintf(buffer, "end   time: %lu.%03lu\n", s2, ms2);
	bfont_draw_str(vram_s+114*320+5, 320, 0, buffer);
	sprintf(buffer, "diff(ms) = %ld\n", s2*1000 + ms2 - s * 1000 - ms);
	bfont_draw_str(vram_s+139*320+5, 320, 0, buffer);
	sprintf(buffer, "hz = %f\n", 500.0 / ((s2*1000 + ms2 - s * 1000 - ms) / 1000.0));
	bfont_draw_str(vram_s+164*320+5, 320, 0, buffer);
	wait_a();
	cls(320,240);
}
#endif /* ifdef HZ_TEST */

extern tUIDriver dc_ui_driver;

KOS_INIT_FLAGS(INIT_IRQ /*| INIT_THD_PREEMPT*/);

int main(int argc, char **argv)
{
	/* set own user interface variation */
	ui_driver = &dc_ui_driver;

	/* initialize screen updater */
	update_screen_updater();

	/* initialize Atari800 core */
	Atari800_Initialise(&argc, argv);

	/* initialize dc controllers */
	dc_controller_init();

	/* initialize sound */
	dc_sound_init();

	/*
	ser_console_init();
	dbgio_init();
	*/

#ifdef MY_RELEASE
	cls(320,240);
	bfont_draw_str(vram_s+50*320+5, 320, 0, "Hallo Tobias und Dominik!");
	bfont_draw_str(vram_s+57*320+5, 320, 0, "_________________________");
	wait_a();
	cls(320,240);
#endif

#ifdef DEBUG
	printf("\nFor Tobias & Dominik\n");
	printf("--------------------\n");
#endif

#ifdef HZ_TEST
	do_hz_test();
#endif

#if 0
	gdb_init();
	asm("mov #7,r12");
	asm("trapa #0x20");
	asm("nop");
	asm("nop");
	asm("nop");
	asm("nop");
	asm("nop");
	asm("nop");
	//gdb_breakpoint();
#endif

	atari_screen_backup = malloc(ATARI_HEIGHT * ATARI_WIDTH);
	atari_screen_backup2 = malloc(ATARI_HEIGHT * ATARI_WIDTH);
	setup_state_dir();

	autostart();

	/* main loop */
	while(TRUE)
	{
		int keycode;

		keycode = Atari_Keyboard();

		switch (keycode) {
		case AKEY_5200_RESET:
			if (machine_type == MACHINE_5200) {
				Coldstart();
			}
			break;
		case AKEY_COLDSTART:
			Coldstart();
			break;
		case AKEY_WARMSTART:
			Warmstart();
			break;
		case AKEY_EXIT:
			Atari800_Exit(FALSE);
			arch_reboot();
#ifdef KB_UI
		case AKEY_KEYB:
			if (machine_type != MACHINE_5200) {
				Sound_Pause();
				if (x_ovr || y_ovr || b_ovr) {
					kb_ui((UBYTE *)atari_screen, NULL, KB_CONSOL);
				}
				else {
					kb_ui((UBYTE *)atari_screen, NULL, 0);
				}
				key_consol |= CONSOL_START;
				/*b_ui_leave = TRUE;  crashes when included!! why?? */
				Sound_Continue();
				controller_update();
			}
			else {
				Sound_Pause();
				kb_ui_5200((UBYTE *)atari_screen);
				Sound_Continue();
				controller_update();
			}
			break;
#endif /* #ifdef KB_UI */
		case AKEY_UI:
			Sound_Pause();
			ui((UBYTE *)atari_screen);
			Sound_Continue();
			controller_update();
			break;
		case AKEY_BREAK:
			key_break = 1;
			break;
		default:
			key_break = 0;
			key_code = keycode;
			break;
		}

		Atari800_Frame(EMULATE_FULL);
		Atari_DisplayScreen((UBYTE *)atari_screen);
		if (! db_mode) dc_atari_sync();  /* with db_mode Atari_DisplayScreen syncs */
		controller_update();  /* get new values from the controllers */
	}
}

/*
 * autostart: check for autorun.xxx files on the
 * CD root directory and start the first one found
 */
static void autostart(void)
{
	struct stat sb;

	if (! stat("/cd/autorun.com", &sb)) {
		if (sb.type == STAT_TYPE_FILE) {
			if (BIN_loader("/cd/autorun.com")) {
				Coldstart();
				return;
			}
		}
	}
	if (! stat("/cd/autorun.exe", &sb)) {
		if (sb.type == STAT_TYPE_FILE) {
			if (BIN_loader("/cd/autorun.exe")) {
				Coldstart();
				return;
			}
		}
	}
	if (! stat("/cd/autorun.atr", &sb)) {
		if (sb.type == STAT_TYPE_FILE) {
			SIO_Mount(1, "/cd/autorun.atr", TRUE);
			return;
		}
	}
}

int sscanf(const char *str, const char *format, ...)
{
	va_list args;
	int *res;

	if (strcmp(format, "%d")) return(0);   /* sanity check, we only support "%d" */

	va_start(args, format);
	res = va_arg(args, int *);
	*res = atoi(str);
	va_end(args);
	return(1);
}

static char mytmpnam[] = "tempname.oef";
char *tmpnam(char *space)
{
	if (space) {
		strcpy(space, mytmpnam);
		return(space);
	}
	else {
		return(mytmpnam);
	}
}
FILE *tmpfile(void) { return(NULL); }
int system(const char* path) { return 0; }

static struct tm localtime_tm;
struct tm *localtime(const time_t *clock)
{
	/* Get the time */
	localtime_r(clock, &localtime_tm);
	return(&localtime_tm);
}

static int odc;
DIR *opendir(const char *name)
{
	static file_t handle;
#ifdef DEBUG
	printf("opendir...(%s)\n", name);
#endif
	if (open_tray) return(NULL);
	/*if (tray_closed)*/ tray_closed = FALSE;
#if 0
	if (! strcmp(name, "/cd")) {
		/* hack */
		cdrom_init();
		iso_ioctl(0, NULL, 0); /* @@@ check what this is in new KOS version ! */
	}
#endif
	if (strcmp(name, "/") && strcmp(name, "/cd") && strcmp(name, "/pc") && strcmp(name, "/ram")
	    && strcmp(name, "/vmu") && strncmp(name, "/cd/", 4) && strncmp(name, "/pc/", 4)
	    && strncmp(name, "/ram/", 5) && strncmp(name, "/vmu/", 5)) {
#ifdef DEBUG
		printf("opendir: punt!\n");
#endif
		return(NULL);
	}
	handle = fs_open(name, O_DIR | O_RDONLY);
	if (! handle) return(NULL);
	if (strcmp(name, "/") && strcmp(name, "/pc")	/* no ".." in the root directory */
	    && strncmp(name, "/pc/", 4))		/* or /pc and subdirectories */
		odc = 1;
	return(&handle);
}

static struct dirent myd;

struct dirent *readdir(DIR *dirp)
{
	if (open_tray) return(NULL);
	if (tray_closed) {
		tray_closed = FALSE;
		return(NULL);
	}
	if (odc) {
		odc = 0;
		memset(&myd, 0, sizeof(struct dirent));
		strcpy(myd.d_name, "..");
		return(&myd);
	}
	return(fs_readdir(*dirp));
}

int closedir(DIR *dirp)
{
	fs_close(*dirp);
	return(0);
}

/* our super-duper stat */
/* only check, whether the S_IFDIR bit has to be set */
int stat(const char *path, struct stat *sb)
{
	file_t handle;

	memset(sb, 0, sizeof(struct stat));  /* preinitialize result */

	if (open_tray) return(-1);
	if (tray_closed) tray_closed = FALSE;

	/* special ".." case */
	if (strlen(path) > 3 &&
	    ! strcmp(path + strlen(path) - 3, "/..")) {
		sb->type = STAT_TYPE_DIR;
		return(0);  /* success */
	}

	/*
	dbgio_printf("stat: path = %s\n", path);
	dbgio_write_str(path); dbgio_write_str("\n");
	*/

#if NOT_YET_WORKING_STAT
	if (fs_stat(path, sb) == -1) {
		printf("stat: error with '%s'\n", path);
		return(-1);
	}
#else
	/* check if dir */
	if ((handle = fs_open(path, O_DIR | O_RDONLY))) {  /* is dir */
		sb->type = STAT_TYPE_DIR;
	}
	else if ((handle = fs_open(path, O_RDONLY))) {	/* is file */
		sb->type = STAT_TYPE_FILE;
	}
	else {	/* is not here */
#ifdef DEBUG
		printf("stat: error with '%s'\n", path);
#endif
		return(-1);
	}
	fs_close(handle);
#endif
	return(0);  /* success */
}

int getchar(void) { return EOF; }
int unlink(const char *path) { return(fs_unlink(path)); }
int remove(const char *path) { return(unlink(path)); }
void dc_perror(const char *x) { printf("perror: '%s'\n", x); }
void (*signal(int sig, void (*func)(int)))(int) { return (void *) 0; }

static void dc_atari_sync(void)
{
	static unsigned long long nextclock = 0;
	unsigned long long curclock;
	uint32 s, ms;

	do {
		timer_ms_gettime(&s, &ms);
		curclock = ((unsigned long long)s << 32) + ms;
	} while (curclock < nextclock);

	nextclock = curclock + CLK_TCK / ((tv_mode == TV_PAL ? 50 : 60));
}

static int sound_enabled = TRUE;	/* sound: on or off */

#define FRAGSIZE       12		/* 1<<FRAGSIZE is size of sound buffer */

void dc_sound_init(void)
{
	snd_init();
#ifdef STEREO
	Pokey_sound_init(FREQ_17_EXACT, DSPRATE, 2, 0, 9);
#else
	Pokey_sound_init(FREQ_17_EXACT, DSPRATE, 1, 0, 0);
#endif
}

void Sound_Pause(void)
{
	if (sound_enabled) {
		sound_enabled = FALSE;
		snd_stream_stop();
	}
}

void Sound_Continue(void)
{
	if (! sound_enabled && glob_snd_ena) {
		sound_enabled = TRUE;
		su_first_call = TRUE;
	}
}

void Sound_Exit(void)
{
#ifdef DEBUG
	printf("Sound_Exit called\n");
#endif
}

/* taken from KOS' snd_stream.c */
#define SPU_RAM_BASE		0xa0800000

#define FRAG_SIZE    0x80      /* size of one fragment */
#define FRAG_NUM     16	       /* max. # of fragments in the buffer */
#define FRAG_BUFSZ   (FRAG_NUM * FRAG_SIZE)

static unsigned char fragbuf[FRAG_SIZE]; /* scratch buffer to generate sound data for a fragment */
static uint32 spu_ram_sch1 = 0;

/* ripped from KOS */
void snd_stream_stop()
{
	AICA_CMDSTR_CHANNEL(tmp, cmd, chan);

	/* Stop stream */
	/* Channel 0 */
	cmd->cmd = AICA_CMD_CHAN;
	cmd->timestamp = 0;
	cmd->size = AICA_CMDSTR_CHANNEL_SIZE;
	cmd->cmd_id = 0;
	chan->cmd = AICA_CH_CMD_STOP;
	snd_sh4_to_aica(tmp, cmd->size);
}

/* ripped from KOS */
static inline unsigned int aica_getpos(int chan)
{
	return(g2_read_32(SPU_RAM_BASE + AICA_CHANNEL(chan) + offsetof(aica_channel_t, pos)));
}

/* some parts taken from KOS' snd_stream.c */
void Sound_Update(void)
{
	static int last_frag;
	int cur_pos, cur_frag, fill_frag, first_frag_to_fill, last_frag_to_fill;

	if (! sound_enabled) return;

	if (su_first_call) {
		AICA_CMDSTR_CHANNEL(tmp, cmd, chan);

		su_first_call = FALSE;

		if (! spu_ram_sch1)
			spu_ram_sch1 = snd_mem_malloc(FRAG_BUFSZ);

		/* prefill buffers */
		Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(spu_ram_sch1, fragbuf, FRAG_SIZE);

		last_frag = 0;

#ifdef STEREO
#error STEREO not implemented yet!
#endif
		/* start streaming */
		/* use channel 0 */
		cmd->cmd = AICA_CMD_CHAN;
		cmd->timestamp = 0;
		cmd->size = AICA_CMDSTR_CHANNEL_SIZE;
		cmd->cmd_id = 0;
		chan->cmd = AICA_CH_CMD_START;
		chan->base = spu_ram_sch1;
		chan->type = AICA_SM_8BIT;
		chan->length = FRAG_BUFSZ;
		chan->loop = 1;
		chan->loopstart = 0;
		chan->loopend = FRAG_BUFSZ - 1;
		chan->freq = DSPRATE;
		chan->vol = 240;
		chan->pan = 127;
		snd_sh4_to_aica(tmp, cmd->size);
	}

	/* get current playing position */
	cur_pos = aica_getpos(0);
	cur_frag = cur_pos / FRAG_SIZE;

	/* calc # of new fragments needed to fill sound buffer */
	first_frag_to_fill = last_frag + 1;
	first_frag_to_fill %= FRAG_NUM;
	last_frag_to_fill = cur_frag;

	fill_frag = first_frag_to_fill;
	while (fill_frag != last_frag_to_fill) {
		Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(spu_ram_sch1 + FRAG_SIZE * fill_frag, fragbuf, FRAG_SIZE);
		last_frag = fill_frag;
		fill_frag++;
		fill_frag %= FRAG_NUM;
	}
#if 0  /* some arm debug stuff */
	printf("cur_pos was %x (%x) (%x)\n", cur_pos,
	       g2_read_32(SPU_RAM_BASE + 0x1f800),
	       g2_read_32(SPU_RAM_BASE + 0x1f804));
#endif
}


static int check_tray_open(void)
{
	int status, disk_type, retval;

	retval = cdrom_get_status(&status, &disk_type);
	if (retval == ERR_OK) {
		if ((status & 15) == 6) return(TRUE);
		return(FALSE);
	}
	else {
		return(FALSE); /* error case: assume tray not open */
	}
}

void dc_strupr(char *s)
{
	char c;
	while ((c = *s)) {
		if (c >= 'a' && c <= 'z') *s = c - 'a' + 'A';
		s++;
	}
}

extern void dc_printbox(UBYTE *screen, char *string);

void dc_please_wait(void)
{
	dc_printbox(atari_screen, " Please wait... ");
	entire_screen_dirty();
	Atari_DisplayScreen(atari_screen);
}

void dc_save_screen(void)
{
	memcpy(atari_screen_backup2, atari_screen, ATARI_HEIGHT * ATARI_WIDTH);
}

void dc_restore_screen(void)
{
	memcpy(atari_screen, atari_screen_backup2, ATARI_HEIGHT * ATARI_WIDTH);
	entire_screen_dirty();
	Atari_DisplayScreen(atari_screen);
}

void dc_error_msg(void)
{
	int key;

#ifdef DEBUG
	printf("dc_error_msg\n");
#endif
	dc_printbox(atari_screen, "     Error!!    ");
	entire_screen_dirty();
	Atari_DisplayScreen(atari_screen);
	key = GetKeyPress(atari_screen);
}

void DCStateSave(void)
{
	unsigned int i, f = 0;

	SaveINT(&screen_tv_mode, 1);
	SaveINT(&tv_mode, 1);
	SaveINT(&x_ovr, 1);
	SaveINT(&x_key, 1);
	SaveINT(&y_ovr, 1);
	SaveINT(&y_key, 1);
	SaveINT(&b_ovr, 1);
	SaveINT(&b_key, 1);
	SaveINT(&emulate_paddles, 1);
	SaveINT(&glob_snd_ena, 1);
	SaveINT(&db_mode, 1);
	SaveINT(&joy_autofire[0], 1);
	for (i=0; i<16; i++) SaveINT(&f, 1);  /* future stuff */

#ifdef DEBUG
        printf("DCStateSave: tv_mode = %d, screen_tv_mode = %d, db_mode = %d\n",
               tv_mode, screen_tv_mode, db_mode);
#endif
}

void DCStateRead(void)
{
	ReadINT(&screen_tv_mode, 1);
	ReadINT(&tv_mode, 1);
	ReadINT(&x_ovr, 1);
	ReadINT(&x_key, 1);
	ReadINT(&y_ovr, 1);
	ReadINT(&y_key, 1);
	ReadINT(&b_ovr, 1);
	ReadINT(&b_key, 1);
	ReadINT(&emulate_paddles, 1);
	ReadINT(&glob_snd_ena, 1);
	ReadINT(&db_mode, 1);
	ReadINT(&joy_autofire[0], 1);
#ifdef DEBUG
        printf("DCStateRead: tv_mode = %d, screen_tv_mode = %d, db_mode = %d\n",
               tv_mode, screen_tv_mode, db_mode);
#endif

	/* some sanity checks */
	if (tv_mode != screen_tv_mode) db_mode = FALSE;
	update_vidmode();
	update_screen_updater();
}

